def getrlimit(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()


def getrusage(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()


def setrlimit(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()
