# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tracking.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    struct_pb2 as google_dot_protobuf_dot_struct__pb2,
    symbol_database as _symbol_database,
    timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2,
)


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="tracking.proto",
    package="dbnd.services.tracking",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0etracking.proto\x12\x16\x64\x62nd.services.tracking\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xe5\x01\n\x05\x45vent\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06schema\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x39\n\x06labels\x18\x04 \x03(\x0b\x32).dbnd.services.tracking.Event.LabelsEntry\x12-\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"q\n\x11PostEventsRequest\x12-\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1d.dbnd.services.tracking.Event\x12-\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"S\n\x12PostEventsResponse\x12*\n\tresponses\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x11\n\texception\x18\x02 \x01(\t"s\n\x12SearchEventRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12\x16\n\tpage_size\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x18\n\x0bpage_number\x18\x03 \x01(\rH\x01\x88\x01\x01\x42\x0c\n\n_page_sizeB\x0e\n\x0c_page_number"D\n\x13SearchEventResponse\x12-\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1d.dbnd.services.tracking.Event2\xc8\x02\n\x0fTrackingService\x12\x63\n\nPostEvents\x12).dbnd.services.tracking.PostEventsRequest\x1a*.dbnd.services.tracking.PostEventsResponse\x12g\n\x0cStreamEvents\x12).dbnd.services.tracking.PostEventsRequest\x1a*.dbnd.services.tracking.PostEventsResponse(\x01\x12g\n\x0cSearchEvents\x12*.dbnd.services.tracking.SearchEventRequest\x1a+.dbnd.services.tracking.SearchEventResponseb\x06proto3',
    dependencies=[
        google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,
    ],
)


_EVENT_LABELSENTRY = _descriptor.Descriptor(
    name="LabelsEntry",
    full_name="dbnd.services.tracking.Event.LabelsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="dbnd.services.tracking.Event.LabelsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="dbnd.services.tracking.Event.LabelsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=290,
    serialized_end=335,
)

_EVENT = _descriptor.Descriptor(
    name="Event",
    full_name="dbnd.services.tracking.Event",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="uuid",
            full_name="dbnd.services.tracking.Event.uuid",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="schema",
            full_name="dbnd.services.tracking.Event.schema",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="dbnd.services.tracking.Event.data",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="labels",
            full_name="dbnd.services.tracking.Event.labels",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="dbnd.services.tracking.Event.timestamp",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[_EVENT_LABELSENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=106,
    serialized_end=335,
)


_POSTEVENTSREQUEST = _descriptor.Descriptor(
    name="PostEventsRequest",
    full_name="dbnd.services.tracking.PostEventsRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="events",
            full_name="dbnd.services.tracking.PostEventsRequest.events",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="dbnd.services.tracking.PostEventsRequest.timestamp",
            index=1,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=337,
    serialized_end=450,
)


_POSTEVENTSRESPONSE = _descriptor.Descriptor(
    name="PostEventsResponse",
    full_name="dbnd.services.tracking.PostEventsResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="responses",
            full_name="dbnd.services.tracking.PostEventsResponse.responses",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="exception",
            full_name="dbnd.services.tracking.PostEventsResponse.exception",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=452,
    serialized_end=535,
)


_SEARCHEVENTREQUEST = _descriptor.Descriptor(
    name="SearchEventRequest",
    full_name="dbnd.services.tracking.SearchEventRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="query",
            full_name="dbnd.services.tracking.SearchEventRequest.query",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="page_size",
            full_name="dbnd.services.tracking.SearchEventRequest.page_size",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="page_number",
            full_name="dbnd.services.tracking.SearchEventRequest.page_number",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_page_size",
            full_name="dbnd.services.tracking.SearchEventRequest._page_size",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_page_number",
            full_name="dbnd.services.tracking.SearchEventRequest._page_number",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=537,
    serialized_end=652,
)


_SEARCHEVENTRESPONSE = _descriptor.Descriptor(
    name="SearchEventResponse",
    full_name="dbnd.services.tracking.SearchEventResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="events",
            full_name="dbnd.services.tracking.SearchEventResponse.events",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=654,
    serialized_end=722,
)

_EVENT_LABELSENTRY.containing_type = _EVENT
_EVENT.fields_by_name["data"].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EVENT.fields_by_name["labels"].message_type = _EVENT_LABELSENTRY
_EVENT.fields_by_name[
    "timestamp"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_POSTEVENTSREQUEST.fields_by_name["events"].message_type = _EVENT
_POSTEVENTSREQUEST.fields_by_name[
    "timestamp"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_POSTEVENTSRESPONSE.fields_by_name[
    "responses"
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_SEARCHEVENTREQUEST.oneofs_by_name["_page_size"].fields.append(
    _SEARCHEVENTREQUEST.fields_by_name["page_size"]
)
_SEARCHEVENTREQUEST.fields_by_name[
    "page_size"
].containing_oneof = _SEARCHEVENTREQUEST.oneofs_by_name["_page_size"]
_SEARCHEVENTREQUEST.oneofs_by_name["_page_number"].fields.append(
    _SEARCHEVENTREQUEST.fields_by_name["page_number"]
)
_SEARCHEVENTREQUEST.fields_by_name[
    "page_number"
].containing_oneof = _SEARCHEVENTREQUEST.oneofs_by_name["_page_number"]
_SEARCHEVENTRESPONSE.fields_by_name["events"].message_type = _EVENT
DESCRIPTOR.message_types_by_name["Event"] = _EVENT
DESCRIPTOR.message_types_by_name["PostEventsRequest"] = _POSTEVENTSREQUEST
DESCRIPTOR.message_types_by_name["PostEventsResponse"] = _POSTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name["SearchEventRequest"] = _SEARCHEVENTREQUEST
DESCRIPTOR.message_types_by_name["SearchEventResponse"] = _SEARCHEVENTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType(
    "Event",
    (_message.Message,),
    {
        "LabelsEntry": _reflection.GeneratedProtocolMessageType(
            "LabelsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _EVENT_LABELSENTRY,
                "__module__": "tracking_pb2"
                # @@protoc_insertion_point(class_scope:dbnd.services.tracking.Event.LabelsEntry)
            },
        ),
        "DESCRIPTOR": _EVENT,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.Event)
    },
)
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.LabelsEntry)

PostEventsRequest = _reflection.GeneratedProtocolMessageType(
    "PostEventsRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTEVENTSREQUEST,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.PostEventsRequest)
    },
)
_sym_db.RegisterMessage(PostEventsRequest)

PostEventsResponse = _reflection.GeneratedProtocolMessageType(
    "PostEventsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTEVENTSRESPONSE,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.PostEventsResponse)
    },
)
_sym_db.RegisterMessage(PostEventsResponse)

SearchEventRequest = _reflection.GeneratedProtocolMessageType(
    "SearchEventRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHEVENTREQUEST,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.SearchEventRequest)
    },
)
_sym_db.RegisterMessage(SearchEventRequest)

SearchEventResponse = _reflection.GeneratedProtocolMessageType(
    "SearchEventResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHEVENTRESPONSE,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.SearchEventResponse)
    },
)
_sym_db.RegisterMessage(SearchEventResponse)


_EVENT_LABELSENTRY._options = None

_TRACKINGSERVICE = _descriptor.ServiceDescriptor(
    name="TrackingService",
    full_name="dbnd.services.tracking.TrackingService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=725,
    serialized_end=1053,
    methods=[
        _descriptor.MethodDescriptor(
            name="PostEvents",
            full_name="dbnd.services.tracking.TrackingService.PostEvents",
            index=0,
            containing_service=None,
            input_type=_POSTEVENTSREQUEST,
            output_type=_POSTEVENTSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="StreamEvents",
            full_name="dbnd.services.tracking.TrackingService.StreamEvents",
            index=1,
            containing_service=None,
            input_type=_POSTEVENTSREQUEST,
            output_type=_POSTEVENTSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SearchEvents",
            full_name="dbnd.services.tracking.TrackingService.SearchEvents",
            index=2,
            containing_service=None,
            input_type=_SEARCHEVENTREQUEST,
            output_type=_SEARCHEVENTRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_TRACKINGSERVICE)

DESCRIPTOR.services_by_name["TrackingService"] = _TRACKINGSERVICE

# @@protoc_insertion_point(module_scope)
